/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

package org.apache.poi.xssf.streaming;

import java.util.Map;
import java.util.TreeMap;

import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.DataFormatter;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.util.SheetUtil;

/**
 * Tracks maximum column width for rows of an {@link SXSSFSheet},
 * to be able to correctly calculate auto-sized column widths even
 * if some rows are already flushed to disk.
*/
/*package-protected*/ class AutoSizeColumnTracker {
    private final SXSSFSheet sheet;
    private int defaultCharWidth = -1;

    private final Map<Integer, ColumnWidthPair> maxColumnWidths = new TreeMap<Integer, ColumnWidthPair>();
    private DataFormatter dataFormatter;

    private static class ColumnWidthPair {
        private double withSkipMergedCells = -1.0;
        private double withUseMergedCells = -1.0;
        
        ColumnWidthPair(final double columnWidthSkipMergedCells, final double columnWidthUseMergedCells) {
            this.withSkipMergedCells = columnWidthSkipMergedCells;
            this.withUseMergedCells = columnWidthUseMergedCells;
        }

        void updateWithMax(final double cellWidthSkipMergedCells, final double cellWidthUseMergedCells) {
            withSkipMergedCells = Math.max(withSkipMergedCells, cellWidthSkipMergedCells);
            withUseMergedCells = Math.max(withUseMergedCells, cellWidthUseMergedCells);
        }
        
        double getMaxColumnWidth(final boolean useMergedCells) {
            return useMergedCells ? withUseMergedCells : withSkipMergedCells;
        }
    }
    
    AutoSizeColumnTracker(final SXSSFSheet sheet) {
        this.sheet = sheet;
    }
    
    void trackCurrentWindowRows() {
        for (int rowIdx = sheet.getFirstRowNum(); rowIdx <= sheet.getLastRowNum(); rowIdx++) {
            Row row = sheet.getRow(rowIdx);
            if (row != null) {
                trackRow(row);
            }
        }
    }

    void trackRow(final Row row) {
        final int firstCellNum = row.getFirstCellNum();
        if (firstCellNum < 0) {
            return;
        }
        final int lastCellNum = row.getLastCellNum();
        
        if (defaultCharWidth < 0) {
            defaultCharWidth = SheetUtil.getDefaultCharWidth(sheet.getWorkbook());
        }

        if (dataFormatter == null) {
            dataFormatter = new DataFormatter();
        }

        for (int cellNum = firstCellNum; cellNum < lastCellNum; cellNum++) {
            final Cell cell = row.getCell(cellNum);
            if (cell == null) {
                continue;
            }
            final double cellWidthSkipMergedCells = SheetUtil.getCellWidth(cell, defaultCharWidth, dataFormatter, false);
            final double cellWidthUseMergedCells = SheetUtil.getCellWidth(cell, defaultCharWidth, dataFormatter, true);
            
            final ColumnWidthPair maxColumnWidthPair = maxColumnWidths.get(cellNum);
            if (maxColumnWidthPair == null) {
                final ColumnWidthPair columnWidthPair = new ColumnWidthPair(cellWidthSkipMergedCells, cellWidthUseMergedCells);
                maxColumnWidths.put(cellNum, columnWidthPair);
            } else {
                maxColumnWidthPair.updateWithMax(cellWidthSkipMergedCells, cellWidthUseMergedCells);
            }
        }
    }

    void autoSizeColumn(final int columnIndex) {
        autoSizeColumn(columnIndex, false);
    }
    
    void autoSizeColumn(final int columnIndex, final boolean useMergedCells) {
        Integer columnWidth = calculateWidthForColumn(columnIndex, useMergedCells);
        if (columnWidth != null) {
            sheet.setColumnWidth(columnIndex, columnWidth);
        }
    }
    
    private Integer calculateWidthForColumn(final int columnIndex, final boolean useMergedCells) {
        final ColumnWidthPair maxColumnWidthPair = maxColumnWidths.get(columnIndex);
        if (maxColumnWidthPair == null) {
            return null;
        }
        double maxColumnWidth = maxColumnWidthPair.getMaxColumnWidth(useMergedCells);
        if (maxColumnWidth > 0.0) {
            maxColumnWidth = Math.min(maxColumnWidth, 255); // The maximum column width for an individual cell is 255 characters
            return (int)(maxColumnWidth*256);
        }
        return null;
    }

}
