package org.apache.poi.xssf.streaming;

import static org.junit.Assert.assertTrue;

import java.util.Arrays;
import java.util.Collection;

import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.Font;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.util.CellRangeAddress;
import org.apache.poi.ss.util.SheetUtil;
import org.junit.After;
import org.junit.Assume;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameter;
import org.junit.runners.Parameterized.Parameters;


/**
 * Tests the auto-sizing behaviour of {@link SXSSFSheet} when not all
 * rows fit into the memory window size etc.
 * 
 * @see Bug #57450 which reported the original mis-behaviour
 */
@RunWith(Parameterized.class)
public class TestAutoSizeColumnTracker {
    
    private static final String SHORT_CELL_VALUE = "Ben";
    private static final String LONG_CELL_VALUE = "B Be Ben Beni Benif Benify Benif Beni Ben Be B";
    
    // Approximative threshold to decide whether test is PASS or FAIL:
    //  shortCellValue ends up with approx column width 1_000 (on my machine),
    //  longCellValue ends up with approx. column width 10_000 (on my machine)
    //  so shortCellValue can be expected to be < 5000 for all fonts
    //  and longCellValue can be expected to be > 5000 for all fonts
    private static final int COLUMN_WIDTH_THRESHOLD_BETWEEN_SHORT_AND_LONG = 5000;
    
    
    private SXSSFSheet sheet;
    private SXSSFWorkbook workbook;
    
    @Parameter(0)
    public boolean useMergedCells;
    
    @Parameters(name="{index}: useMergedCells={0}")
    public static Collection<Object[]> data() {
        return Arrays.asList(new Object[][] {     
                 {false},
                 {true}, 
           });
    }
    
    @After
    public void tearDownSheetAndWorkbook() throws Exception {
        if (sheet != null) {
            sheet.dispose();
        }
        if (workbook != null) {
            workbook.close();
        }
    }
    
    @Test
    public void test_EmptySheet_NoException() {
        workbook = new SXSSFWorkbook();
        sheet = workbook.createSheet();
        
        for (int i = 0; i < 10; i++) {
            sheet.autoSizeColumn(i, useMergedCells);
        }
    }
    
    @Test
    public void test_WindowSizeDefault_AllRowsFitIntoWindowSize() throws Exception {
        workbook = new SXSSFWorkbook();
        sheet = workbook.createSheet();
        
        final Cell cellRow0 = createRowWithCellValues(sheet, 0, LONG_CELL_VALUE);
        
        assumeRequiredFontsAreInstalled(workbook, cellRow0);

        createRowWithCellValues(sheet, 1, SHORT_CELL_VALUE);

        sheet.autoSizeColumn(0, useMergedCells);

        assertColumnWidthStrictlyWithinRange(sheet.getColumnWidth(0), COLUMN_WIDTH_THRESHOLD_BETWEEN_SHORT_AND_LONG, 255*256);
    }
    
    @Test
    public void test_WindowSizeEqualsOne_IgnoreFlushedRows() throws Exception {
        workbook = new SXSSFWorkbook(null, 1); // Window size 1 so only last row will be in memory
        sheet = workbook.createSheet();
        
        final Cell cellRow0 = createRowWithCellValues(sheet, 0, LONG_CELL_VALUE);
        
        assumeRequiredFontsAreInstalled(workbook, cellRow0);

        createRowWithCellValues(sheet, 1, SHORT_CELL_VALUE);

        sheet.autoSizeColumn(0, useMergedCells);

        assertColumnWidthStrictlyWithinRange(sheet.getColumnWidth(0), 0, COLUMN_WIDTH_THRESHOLD_BETWEEN_SHORT_AND_LONG);
    }

    @Test
    public void test_WindowSizeEqualsOne_lastRowIsNotWidest() throws Exception {
        workbook = new SXSSFWorkbook(null, 1); // Window size 1 so only last row will be in memory
        sheet = workbook.createSheet();
        sheet.setAutoSizeTrackAllFlushedRows(true);
        
        final Cell cellRow0 = createRowWithCellValues(sheet, 0, LONG_CELL_VALUE);
        
        assumeRequiredFontsAreInstalled(workbook, cellRow0);

        createRowWithCellValues(sheet, 1, SHORT_CELL_VALUE);

        sheet.autoSizeColumn(0, useMergedCells);

        assertColumnWidthStrictlyWithinRange(sheet.getColumnWidth(0), COLUMN_WIDTH_THRESHOLD_BETWEEN_SHORT_AND_LONG, 255*256);
    }
    
    @Test
    public void test_WindowSizeEqualsOne_lastRowIsWidest() throws Exception {
        workbook = new SXSSFWorkbook(null, 1); // Window size 1 so only last row will be in memory
        sheet = workbook.createSheet();
        sheet.setAutoSizeTrackAllFlushedRows(true);
        
        final Cell cellRow0 = createRowWithCellValues(sheet, 0, SHORT_CELL_VALUE);

        assumeRequiredFontsAreInstalled(workbook, cellRow0);
        
        createRowWithCellValues(sheet, 1, LONG_CELL_VALUE);

        sheet.autoSizeColumn(0, useMergedCells);

        assertColumnWidthStrictlyWithinRange(sheet.getColumnWidth(0), COLUMN_WIDTH_THRESHOLD_BETWEEN_SHORT_AND_LONG, 255*256);
    }
    
    @Test
    public void test_WindowSizeEqualsOne_flushedRowHasMergedCell() throws Exception {
        workbook = new SXSSFWorkbook(null, 1); // Window size 1 so only last row will be in memory
        sheet = workbook.createSheet();
        sheet.setAutoSizeTrackAllFlushedRows(true);
        
        Cell a1 = createRowWithCellValues(sheet, 0, LONG_CELL_VALUE);

        assumeRequiredFontsAreInstalled(workbook, a1);
        sheet.addMergedRegion(new CellRangeAddress(0, 0, 0, 1)); // merge cells A1 and A2
        
        createRowWithCellValues(sheet, 1, SHORT_CELL_VALUE, SHORT_CELL_VALUE);
        
        sheet.autoSizeColumn(0, useMergedCells);
        sheet.autoSizeColumn(1, useMergedCells);

        if (useMergedCells) {
            assertColumnWidthStrictlyWithinRange(sheet.getColumnWidth(0), COLUMN_WIDTH_THRESHOLD_BETWEEN_SHORT_AND_LONG, 255*256);
        } else {
            assertColumnWidthStrictlyWithinRange(sheet.getColumnWidth(0), 0, COLUMN_WIDTH_THRESHOLD_BETWEEN_SHORT_AND_LONG);
        }
        assertColumnWidthStrictlyWithinRange(sheet.getColumnWidth(1), 0, COLUMN_WIDTH_THRESHOLD_BETWEEN_SHORT_AND_LONG);
    }
    
    private static void assumeRequiredFontsAreInstalled(final Workbook workbook, final Cell cell) {
        // autoSize will fail if required fonts are not installed, skip this test then
        Font font = workbook.getFontAt(cell.getCellStyle().getFontIndex());
        Assume.assumeTrue("Cannot verify autoSizeColumn() because the necessary Fonts are not installed on this machine: " + font,
                          SheetUtil.canComputeColumnWidth(font));
    }
    
    private static Cell createRowWithCellValues(final Sheet sheet, final int rowNumber, final String... cellValues) {
        Row row = sheet.createRow(rowNumber);
        int cellIndex = 0;
        Cell firstCell = null;
        for (final String cellValue : cellValues) {
            Cell cell = row.createCell(cellIndex++);
            if (firstCell == null) {
                firstCell = cell;
            }
            cell.setCellValue(cellValue);
        }
        return firstCell;
    }
    
    private static void assertColumnWidthStrictlyWithinRange(final int actualColumnWidth, final int lowerBoundExclusive, final int upperBoundExclusive) {
        assertTrue("Expected a column width greater than " + lowerBoundExclusive + " but found " + actualColumnWidth,
                actualColumnWidth > lowerBoundExclusive);
        assertTrue("Expected column width less than " + upperBoundExclusive + " but found " + actualColumnWidth, actualColumnWidth < upperBoundExclusive);
       
    }
    
}
